#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-query-dois
# Created by the Natural History Museum in London, UK

import json
import logging
from contextlib import suppress

from ckan import plugins

from . import helpers, routes, cli
from .lib.doi import mint_doi, mint_multisearch_doi, find_existing_doi
from .lib.query import DatastoreQuery
from .lib.stats import DOWNLOAD_ACTION, record_stat
from .logic import auth, action
from .logic.utils import extract_resource_ids_and_versions

log = logging.getLogger(__name__)


class QueryDOIsPlugin(plugins.SingletonPlugin):
    plugins.implements(plugins.IBlueprint, inherit=True)
    plugins.implements(plugins.IConfigurer, inherit=True)
    plugins.implements(plugins.ITemplateHelpers)
    plugins.implements(plugins.IActions)
    plugins.implements(plugins.IAuthFunctions)
    plugins.implements(plugins.IClick)
    # if the ckanpackager is available, we have a hook for it
    with suppress(ImportError):
        from ckanext.ckanpackager.interfaces import ICkanPackager

        plugins.implements(ICkanPackager)
    # if the versioned datastore downloader is available, we have a hook for it
    with suppress(ImportError):
        from ckanext.versioned_datastore.interfaces import IVersionedDatastoreDownloads

        plugins.implements(IVersionedDatastoreDownloads, inherit=True)

    # IBlueprint
    def get_blueprint(self):
        return routes.blueprints

    # IClick
    def get_commands(self):
        return cli.get_commands()

    # IAuthFunctions
    def get_auth_functions(self):
        return {
            'create_doi': auth.create_doi,
        }

    # IActions
    def get_actions(self):
        return {
            'create_doi': action.create_doi,
        }

    # IConfigurer
    def update_config(self, config):
        # add templates
        plugins.toolkit.add_template_directory(config, 'theme/templates')
        # add the resource groups
        plugins.toolkit.add_resource('theme/assets', 'ckanext-query-dois')

    # IVersionedDatastoreDownloads
    def download_modify_notifier_template_context(self, request, context):
        try:
            # if a DOI can be created it should already have been created in download_before_write
            doi = find_existing_doi(
                request.core_record.resource_ids_and_versions,
                request.core_record.query_hash,
                request.core_record.query_version,
            )

            if doi:
                # update the context with the doi
                context['doi'] = doi.doi
        except:
            # if anything goes wrong we don't want to stop the download; just log the
            # error and move on
            log.error('Failed to retrieve DOI and/or create stats', exc_info=True)

        # always return the context
        return context

    def download_modify_manifest(self, manifest, request):
        try:
            # check to see if the download is something we can stick a DOI on (this will
            # throw a validation error if any of the resources aren't valid for DOI-ing
            extract_resource_ids_and_versions(
                req_resource_ids_and_versions=request.core_record.resource_ids_and_versions
            )

            # mint the DOI on datacite if necessary
            created, doi = mint_multisearch_doi(
                request.core_record.query,
                request.core_record.query_version,
                request.core_record.resource_ids_and_versions,
            )

            # add the doi to the manifest
            manifest['query-doi'] = doi.doi
        except:
            # if anything goes wrong we don't want to stop the download from completing;
            # just log the error and move on
            log.error('Failed to mint/retrieve DOI', exc_info=True)

        # always return the manifest
        return manifest

    def download_after_run(self, request):
        try:
            # if a DOI can be created it should already have been created in
            # download_modify_manifest
            doi = find_existing_doi(
                request.core_record.resource_ids_and_versions,
                request.core_record.query_hash,
                request.core_record.query_version,
            )

            if doi and request.state == 'complete':
                # record a download stat against the DOI
                record_stat(doi, DOWNLOAD_ACTION, identifier=request.id)
        except:
            # just log the error and move on
            log.error('Failed to retrieve DOI and/or create stats', exc_info=True)

    # ICkanPackager
    def before_package_request(
        self, resource_id, package_id, packager_url, request_params
    ):
        """
        This is called prior to the ckanpackager plugin starting it's creation of a
        download, therefore we can use it to mint DOIs and record a stat.

        :param resource_id: the resource id of the resource that is about to be packaged
        :param package_id: the package id of the resource that is about to be packaged
        :param packager_url: the target url for this packaging request
        :param request_params: a dict of parameters that will be sent with the request
        :return: the url and the params as a tuple
        """
        resource = plugins.toolkit.get_action('resource_show')({}, {'id': resource_id})
        package = plugins.toolkit.get_action('package_show')({}, {'id': package_id})
        # only handle DOIs for resources with data in the datastore and that aren't in private
        # packages
        if resource.get('datastore_active', False) and not package.get('private', True):
            try:
                # the request_params dict comes through with a lot of other parameters in it beyond
                # the query params, therefore we need to do some stripping out (otherwise the exact
                # same query will look different due to a different format choice or email address)
                allowed_keys = {'q', 'filters', 'version'}
                data_dict = {
                    k: v for k, v in request_params.items() if k in allowed_keys
                }

                # the packager converts the filters dict to JSON so we need to convert it back
                if 'filters' in data_dict:
                    data_dict['filters'] = json.loads(data_dict['filters'])

                # mint the DOI on datacite if necessary
                _minted, query_doi = mint_doi(
                    [resource_id], DatastoreQuery(data_dict=data_dict)
                )
                # record a download stat against the DOI
                record_stat(query_doi, DOWNLOAD_ACTION, request_params['email'])
                # add the doi to the ckanpackager params so that it can be reported in the email
                request_params['doi'] = query_doi.doi
            except:
                # if anything goes wrong we don't want to stop the download from going ahead, just
                # log the error and move on
                log.error(
                    'Failed to mint/retrieve DOI and/or create stats', exc_info=True
                )

        return packager_url, request_params

    # ITemplateHelpers
    def get_helpers(self):
        return {
            'render_filter_value': helpers.render_filter_value,
            'get_most_recent_dois': helpers.get_most_recent_dois,
            'get_time_ago_description': helpers.get_time_ago_description,
            'get_landing_page_url': helpers.get_landing_page_url,
            'create_citation_text': helpers.create_citation_text,
            'create_multisearch_citation_text': helpers.create_multisearch_citation_text,
            'pretty_print_query': helpers.pretty_print_query,
            'get_doi_count': helpers.get_doi_count,
        }



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-query-dois
# Created by the Natural History Museum in London, UK

import logging
import random
import string
from ckan import model
from ckan.common import asbool
from ckan.plugins import toolkit
from datacite import DataCiteMDSClient, schema41
from datacite.errors import DataCiteError, DataCiteNotFoundError
from datetime import datetime

from .utils import get_resource_and_package, get_authors, get_resource_counts
from ..model import QueryDOI

log = logging.getLogger(__name__)

# this is a test prefix available for all minters to use for testing purposes
TEST_PREFIX = '10.5072'


def is_test_mode():
    """
    Should we use the test datacite MDS API?

    :return: True if we should, False if not. Defaults to True.
    """
    return asbool(toolkit.config.get('ckanext.query_dois.test_mode', True))


def get_prefix():
    """
    Gets the prefix to use for the DOIs we mint.

    :return: the prefix to use for the new DOIs
    """
    prefix = toolkit.config.get('ckanext.query_dois.prefix')

    if prefix is None:
        raise TypeError('You must set the ckanext.query_dois.prefix config value')

    if prefix == TEST_PREFIX:
        raise ValueError(
            f'The test prefix {TEST_PREFIX} has been retired, use a prefix defined in '
            f'your datacite test account'
        )

    return prefix


def get_client():
    """
    Get a datacite MDS API client, configured for use.

    :return: a DataCite client object
    """
    kwargs = dict(
        username=toolkit.config.get('ckanext.query_dois.datacite_username'),
        password=toolkit.config.get('ckanext.query_dois.datacite_password'),
        prefix=get_prefix(),
        test_mode=is_test_mode(),
    )
    # datacite 1.0.1 isn't updated for the test prefix deprecation yet so this is a temp fix
    if is_test_mode():
        kwargs.update({'url': 'https://mds.test.datacite.org'})
    return DataCiteMDSClient(**kwargs)


def generate_doi(client):
    """
    Generate a new DOI which isn't currently in use. The database is checked for
    previous usage, as is Datacite itself. Use whatever value is retuned from this
    function quickly to avoid double use as this function uses no locking.

    :param client: an instance of the DataCiteMDSClient class
    :return: the full, unique DOI
    """
    # the list of valid characters is larger than just lowercase and the digits but we don't need
    # that many options and URLs with just alphanumeric characters in them are nicer. We just use
    # lowercase characters to avoid any issues with case being ignored
    valid_characters = string.ascii_lowercase + string.digits

    attempts = 5

    while attempts > 0:
        # generate a random 8 character identifier and prepend qd. to make it easier to identify
        # DOIs from this extension
        identifier = f'qd.{"".join(random.choice(valid_characters) for _ in range(8))}'
        # form the doi using the prefix
        doi = f'{get_prefix()}/{identifier}'

        # check this doi doesn't exist in the table
        if model.Session.query(QueryDOI).filter(QueryDOI.doi == doi).count():
            continue

        # check against the datacite service
        try:
            client.metadata_get(doi)
            # if a doi is found, we need to try again
            continue
        except DataCiteNotFoundError:
            # if no doi is found, we're good!
            pass
        except DataCiteError as e:
            log.warning(
                f'Error whilst checking new DOIs with DataCite. DOI: {doi}, error: {e}'
            )
            attempts -= 1
            continue

        # if we've made it this far the doi isn't in the database and it's not in datacite already
        return doi
    else:
        raise Exception('Failed to generate a DOI')


def find_existing_doi(resources_and_versions, query_hash, query_version=None):
    """
    Returns a QueryDOI object representing the same search, or returns None if one
    doesn't exist.

    :param resources_and_versions: the resource ids being queried mapped to the versions they're
                                   being queried at
    :param query_hash: the hash of the query
    :param query_version: the query version
    :return: a QueryDOI object or None
    """
    return (
        model.Session.query(QueryDOI)
        .filter(
            QueryDOI.resources_and_versions == resources_and_versions,
            QueryDOI.query_hash == query_hash,
            QueryDOI.query_version == query_version,
        )
        .first()
    )


def create_doi_on_datacite(client, doi, authors, timestamp, count):
    """
    Mints the given DOI on datacite using the client.

    :param client: the MDS datacite client
    :param doi: the doi (full, prefix and suffix)
    :param authors: the authors to associate with the DOI
    :param timestamp: the datetime when the DOI was created
    :param count: the number of records contained in the DOI's data
    """
    # create the data for datacite
    data = {
        'identifier': {
            'identifier': doi,
            'identifierType': 'DOI',
        },
        'creators': [{'creatorName': author} for author in authors],
        'titles': [
            {
                'title': toolkit.config.get('ckanext.query_dois.doi_title').format(
                    count=count
                )
            }
        ],
        'publisher': toolkit.config.get('ckanext.query_dois.publisher'),
        'publicationYear': str(timestamp.year),
        'resourceType': {'resourceTypeGeneral': 'Dataset'},
    }

    # use an assert here because the data should be valid every time, otherwise it's something the
    # developer is going to have to fix
    assert schema41.validate(data)

    # create the metadata on datacite
    client.metadata_post(schema41.tostring(data))

    # create the URL the DOI will point to, i.e. the landing page
    data_centre, identifier = doi.split('/')
    landing_page_url = toolkit.url_for(
        'query_doi.landing_page', data_centre=data_centre, identifier=identifier
    )
    site = toolkit.config.get('ckan.site_url')
    if site[-1] == '/':
        site = site[:-1]
    # mint the DOI
    client.doi_post(doi, site + landing_page_url)


def create_database_entry(
    doi,
    query,
    query_hash,
    resources_and_versions,
    timestamp,
    record_count,
    requested_version=None,
    query_version=None,
    resource_counts=None,
):
    """
    Inserts the database row for the query DOI.

    :param doi: the doi (full, prefix and suffix)
    :param query: the query dict
    :param query_hash: the query hash
    :param resources_and_versions: the resource ids mapped to their rounded versions (as a dict)
    :param timestamp: the datetime the DOI was created
    :param record_count: the number of records contained in the DOI's data
    :param requested_version: the version requested by the user, if provided
    :param query_version: the query version, if provided
    :param resource_counts: the resource counts, if provided
    :return: the QueryDOI object
    """
    query_doi = QueryDOI(
        doi=doi,
        resources_and_versions=resources_and_versions,
        timestamp=timestamp,
        query=query,
        query_hash=query_hash,
        requested_version=requested_version,
        count=record_count,
        query_version=query_version,
        resource_counts=resource_counts,
    )
    query_doi.save()
    return query_doi


def mint_doi(resource_ids, datastore_query):
    """
    Mint a DOI on datacite using their API and create a new QueryDOI object, saving it
    to the database. If we already have a query which would produce identical data to
    the one passed then we return the existing QueryDOI object and don't mint or insert
    anything.

    :param resource_ids: a list or resource ids against which the query should be run. Note that as
                         this extension doesn't support multi-resource search yet an exception will
                         be thrown if the length of this list isn't 1.
    :param datastore_query: the DatastoreQuery object containing the query information
    :return: a boolean indicating whether a new DOI was minted and the QueryDOI object representing
             the query's DOI
    """
    # currently we only deal with single resource searches
    if len(resource_ids) != 1:
        raise NotImplemented(
            "This plugin currently doesn't support multi-resource searches"
        )
    resource_id = resource_ids[0]
    rounded_version = datastore_query.get_rounded_version(resource_id)
    resources_and_versions = {resource_id: rounded_version}

    existing_doi = find_existing_doi(resources_and_versions, datastore_query.query_hash)
    if existing_doi is not None:
        return False, existing_doi

    # collect up some details we're going to need to mint the DOI
    timestamp = datetime.now()
    resource, package = get_resource_and_package(resource_id)
    record_count = datastore_query.get_count(resource_id)
    client = get_client()

    # generate a new DOI to store this query against
    doi = generate_doi(client)
    create_doi_on_datacite(client, doi, [package['author']], timestamp, record_count)
    query_doi = create_database_entry(
        doi,
        datastore_query.query,
        datastore_query.query_hash,
        resources_and_versions,
        timestamp,
        record_count,
        requested_version=datastore_query.requested_version,
    )
    return True, query_doi


def mint_multisearch_doi(query, query_version, resource_ids_and_versions):
    """
    Mint a DOI on datacite using their API and create a new QueryDOI object, saving it
    to the database. If we already have a query which would produce identical data to
    the one passed then we return the existing QueryDOI object and don't mint or insert
    anything.

    This function handles DOIs created for the versioned datastore's multisearch action.

    :param query: the query dict
    :param query_version: the query schema version
    :param resource_ids_and_versions: a dict of resource ids -> versions
    :return: a boolean indicating whether a new DOI was minted and the QueryDOI object representing
             the query's DOI
    """
    # first off, ask the versioned datastore extension to create a hash of the query
    hash_data_dict = dict(query=query, query_version=query_version)
    query_hash = toolkit.get_action('datastore_hash_query')({}, hash_data_dict)

    # now check if there are any dois already for this query
    existing_doi = find_existing_doi(
        resource_ids_and_versions, query_hash, query_version
    )
    if existing_doi is not None:
        return False, existing_doi

    # collect up some details we're going to need to mint the DOI
    timestamp = datetime.now()
    authors = get_authors(resource_ids_and_versions.keys())
    # find out how many records match the query
    search_data_dict = {
        'query': query,
        'query_version': query_version,
        'resource_ids_and_versions': resource_ids_and_versions,
        'size': 0,
    }
    record_count = toolkit.get_action('datastore_multisearch')({}, search_data_dict)[
        'total'
    ]
    # find out how many records come from each resource
    resource_counts = get_resource_counts(
        query, query_version, resource_ids_and_versions
    )
    # the list call is used here so that we can modify the resource_ids_and_versions dict as we
    # iterate over it without causing an error
    for resource_id in list(resource_ids_and_versions.keys()):
        if resource_counts[resource_id] == 0:
            del resource_ids_and_versions[resource_id]
            del resource_counts[resource_id]

    if not resource_ids_and_versions:
        raise toolkit.ValidationError(
            'The DOI must be associated with at least one record'
        )

    # generate a new DOI to store this query against
    client = get_client()
    doi = generate_doi(client)
    create_doi_on_datacite(client, doi, authors, timestamp, record_count)
    query_doi = create_database_entry(
        doi,
        query,
        query_hash,
        resource_ids_and_versions,
        timestamp,
        record_count,
        query_version=query_version,
        resource_counts=resource_counts,
    )
    return True, query_doi



# !/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-query-dois
# Created by the Natural History Museum in London, UK


from ckan import model
from ckan.plugins import toolkit
from flask import Blueprint, jsonify

from . import _helpers
from ..model import QueryDOI, QueryDOIStat

blueprint = Blueprint(name='query_doi', import_name=__name__, url_prefix='/doi')


@blueprint.route('/<data_centre>/<identifier>')
def landing_page(data_centre, identifier):
    """
    Renders the landing page for the given DOI.

    :param data_centre: the data centre prefix
    :param identifier: the DOI identifier
    :return: the rendered landing page
    """
    doi = '{}/{}'.format(data_centre, identifier)
    query_doi = _helpers.get_query_doi(doi)
    if query_doi is None:
        raise toolkit.abort(404, toolkit._('DOI not recognised'))

    if query_doi.query_version is not None:
        return _helpers.render_multisearch_doi_page(query_doi)
    else:
        return _helpers.render_datastore_search_doi_page(query_doi)


@blueprint.route('')
def doi_stats():
    """
    Returns statistics in JSON format depending on the request parameters. The return
    will be a list with a dict representing the QueryDOI as each element.

    This endpoint currently only supports filtering on the resource_id.

    :return: a JSON stringified list of dicts
    """
    query = model.Session.query(QueryDOI)

    # by default order by id desc to get the latest first
    query = query.order_by(QueryDOI.id.desc())

    resource_id = toolkit.request.params.get('resource_id', None)
    if resource_id:
        query = query.filter(QueryDOI.on_resource(resource_id))

    # apply the offset and limit, with sensible defaults
    query = query.offset(toolkit.request.params.get('offset', 0))
    query = query.limit(toolkit.request.params.get('limit', 100))

    # return the data as a JSON dumped list of dicts
    return jsonify([stat.as_dict() for stat in query])


@blueprint.route('/stats')
def action_stats():
    """
    Returns action statistics in JSON format depending on the request parameters. The
    return will be a list with a dict representing the QueryDOIStat as each element.

    :return: a JSON stringified list of dicts
    """
    query = model.Session.query(QueryDOIStat)

    # by default order by id desc to get the latest first
    query = query.order_by(QueryDOIStat.id.desc())

    # apply any parameters as filters
    for param_name, column in _helpers.column_param_mapping:
        param_value = toolkit.request.params.get(param_name, None)
        if param_value:
            query = query.filter(column == param_value)

    resource_id = toolkit.request.params.get('resource_id', None)
    if resource_id:
        query = query.join(QueryDOI, QueryDOI.doi == QueryDOIStat.doi).filter(
            QueryDOI.on_resource(resource_id)
        )

    # apply the offset and limit, with sensible defaults
    query = query.offset(toolkit.request.params.get('offset', 0))
    query = query.limit(toolkit.request.params.get('limit', 100))

    # return the data as a JSON dumped list of dicts
    return jsonify([stat.as_dict() for stat in query])



import random
import string
from datetime import datetime

import pytest
from ckan.tests import factories
from ckanext.query_dois.lib.doi import create_database_entry
from ckanext.query_dois.helpers import get_doi_count, get_most_recent_dois
from ckanext.query_dois.model import QueryDOI


def make_doi(resource_id) -> QueryDOI:
    random_str = "".join(random.choice(string.ascii_lowercase) for _ in range(10))
    return create_database_entry(
        doi=random_str,
        query={},
        query_hash=random_str,
        resources_and_versions={resource_id: 4},
        timestamp=datetime.now(),
        record_count=1,
    )


@pytest.mark.usefixtures("clean_db", "setup_db")
class TestGetDOICount:
    def test_when_dois_have_been_made(self):
        # create 2 packages
        package_1 = factories.Dataset(name="package1")
        package_2 = factories.Dataset(name="package2")
        # create 2 resources per package
        resource_1 = factories.Resource(package_id=package_1["id"])
        resource_2 = factories.Resource(package_id=package_1["id"])
        resource_3 = factories.Resource(package_id=package_2["id"])
        resource_4 = factories.Resource(package_id=package_2["id"])

        dois = [
            make_doi(resource_1["id"]),
            make_doi(resource_1["id"]),
            make_doi(resource_2["id"]),
            make_doi(resource_2["id"]),
            make_doi(resource_2["id"]),
            make_doi(resource_3["id"]),
            make_doi(resource_4["id"]),
            make_doi(resource_4["id"]),
            make_doi(resource_4["id"]),
            make_doi(resource_4["id"]),
            make_doi(resource_4["id"]),
            make_doi(resource_4["id"]),
        ]
        assert get_doi_count(package_1["id"]) == 5
        assert get_doi_count(package_2["id"]) == 7

    def test_no_dois_have_been_made(self):
        # create 2 packages
        package_1 = factories.Dataset(name="package1")
        package_2 = factories.Dataset(name="package2")
        # create 2 resources per package
        resource_1 = factories.Resource(package_id=package_1["id"])
        resource_2 = factories.Resource(package_id=package_1["id"])
        resource_3 = factories.Resource(package_id=package_2["id"])
        resource_4 = factories.Resource(package_id=package_2["id"])
        # create a doi on a resource in package 2
        make_doi(resource_3["id"])
        assert get_doi_count(package_1["id"]) == 0

    def test_no_resources(self):
        # create a package
        package_1 = factories.Dataset(name="package1")
        # create no resources
        assert get_doi_count(package_1["id"]) == 0

    def test_package_not_exist(self):
        assert get_doi_count("fiowenfwuibefuiwbefuiwbef") == 0


@pytest.mark.usefixtures("clean_db", "setup_db")
class TestGetMostRecentDOIs:
    def test_no_package(self):
        assert len(get_most_recent_dois("efiownfwe", 5)) == 0

    def test_no_resources(self):
        package_1 = factories.Dataset(name="package1")
        assert len(get_most_recent_dois(package_1["id"], 5)) == 0

    def test_no_dois(self):
        package_1 = factories.Dataset(name="package1")
        resource_1 = factories.Resource(package_id=package_1["id"])
        assert len(get_most_recent_dois(package_1["id"], 5)) == 0

    def test_dois_less_than_limit(self):
        package_1 = factories.Dataset(name="package1")
        resource_1 = factories.Resource(package_id=package_1["id"])
        resource_2 = factories.Resource(package_id=package_1["id"])
        make_doi(resource_1["id"])
        make_doi(resource_2["id"])
        assert len(get_most_recent_dois(package_1["id"], 5)) == 2

    def test_dois_more_than_limit(self):
        package_1 = factories.Dataset(name="package1")
        resource_1 = factories.Resource(package_id=package_1["id"])
        for _ in range(10):
            make_doi(resource_1["id"])
        assert len(get_most_recent_dois(package_1["id"], 5)) == 5



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-query-dois
# Created by the Natural History Museum in London, UK

from ckan.plugins import toolkit

from . import schema as schema_lib
from .utils import extract_resource_ids_and_versions
from ..lib.doi import mint_multisearch_doi
from ..lib.emails import send_saved_search_email
from ..lib.stats import SAVE_ACTION, record_stat


def create_doi(context, data_dict):
    """
    Creates a DOI using the given parameters and returns it.

    Params:

    :param email_address: the email address of the DOI requester
    :type email_address: string
    :param query: the query to associate with the DOI
    :type query: dict
    :param query_version: the query schema version for the query
    :type query_version: string
    :param version: the version to search the data at
    :type version: int, number of milliseconds (not seconds!) since UNIX epoch
    :param resource_ids: the resource ids to search
    :type resource_ids: list of strings
    :param resource_ids_and_versions: a dict of resource ids -> versions defining specific versions
                                      that the resources should be searched at
    :type resource_ids_and_versions: a dict of strings -> version ints

    Returns:

    :rtype: dict
    :param doi: the doi that was created (prefix/suffix)
    :type doi: string
    :param is_new: whether the doi was newly created or whether an existing DOI for the query
                   parameters already existed
    :type is_new: bool
    :param email_sent: whether the email was sent successfully or not
    :type email_sent: bool
    """
    # validate the data dict first
    schema = context.get('schema', schema_lib.create_doi())
    data_dict, errors = toolkit.navl_validate(data_dict, schema, context)
    if errors:
        raise toolkit.ValidationError(errors)

    # then extract the parameters from the data dict, defaulting some things as we go
    email_address = data_dict['email_address']
    query = data_dict.get('query', {})
    query_version = data_dict.get(
        'query_version',
        toolkit.get_action('datastore_get_latest_query_schema_version')({}, {}),
    )
    version = data_dict.get('version', None)
    resource_ids = data_dict.get('resource_ids', None)
    resource_ids_and_versions = data_dict.get('resource_ids_and_versions', None)

    # figure out which resources and which versions we're going to be creating a DOI for
    resource_ids_and_versions = extract_resource_ids_and_versions(
        version, resource_ids, resource_ids_and_versions
    )

    # create a new DOI or retrieve an existing one
    created, doi = mint_multisearch_doi(query, query_version, resource_ids_and_versions)
    # record a stat for this action
    record_stat(doi, SAVE_ACTION, email_address)
    # send the email to the requesting user
    email_sent = send_saved_search_email(email_address, doi)

    return {
        'is_new': created,
        'doi': doi.doi,
        'email_sent': email_sent,
    }



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-query-dois
# Created by the Natural History Museum in London, UK

import copy
import hashlib
import json
import time
from collections import defaultdict

from ckan.plugins import toolkit


class DatastoreQuery(object):
    """
    This models datastore queries passed to datastore_search, not the DOIs created from
    them.
    """

    @staticmethod
    def _parse_from_query_dict(query_dict):
        '''
        Parse a dict of query string parameters which represents the data dict for the
        datastore_search action in the URL format used by CKAN. The query_dict parameter is expected
        to look something like this (for example):

            {
                "q": "banana",
                "filters": "colour:yellow|length:200|colour:brown|type:tasty",
                etc
            }

        If a version is present, either as the version parameter or as the __version__ filter, it
        is extracted with preference given to the version parameter if both are provided.

        :param query_dict: the query string dict
        :return: the query dict (defaults to {} if nothing can be extracted from the query_dict) and
                 the requested version (defaults to None, if not provided in the query_dict)
        '''
        query = {}
        requested_version = None
        for param, param_value in query_dict.items():
            if param == 'version':
                requested_version = int(param_value)
            elif param == 'filters':
                filters = defaultdict(list)
                for filter_pair in param_value.split('|'):
                    filter_field, filter_value = filter_pair.split(':', 1)
                    filters[filter_field].append(filter_value)
                if requested_version is None:
                    popped_version = filters.pop('__version__', None)
                    if popped_version:
                        requested_version = int(popped_version[0])
                if filters:
                    query[param] = filters
            else:
                query[param] = param_value
        return query, requested_version

    @staticmethod
    def _parse_from_data_dict(data_dict):
        '''
        Parse a dict of query string parameters which represents the data dict for the
        datastore_search action in data dict form it expects. The data_dict parameter is expected to
        look something like this (for example):

            {
                "q": "banana",
                "filters": {
                    "colour": ["yellow", "brown"],
                    "length": "200",
                    "type": ["tasty"],
                }
                etc
            }

        If a version is present, either as the version parameter or as the __version__ filter, it
        is extracted with preference given to the version parameter if both are provided.

        :param data_dict: the query string dict
        :return: the query dict (defaults to {} if nothing can be extracted from the query_dict) and
                 the requested version (defaults to None, if not provided in the query_dict)
        '''
        query = {}
        requested_version = None
        for param, param_value in data_dict.items():
            if param == 'version':
                requested_version = int(param_value)
            elif param == 'filters':
                filters = {}
                for filter_field, filter_value in param_value.items():
                    if not isinstance(filter_value, list):
                        filter_value = [filter_value]
                    filters[filter_field] = filter_value
                if requested_version is None:
                    popped_version = filters.pop('__version__', None)
                    if popped_version:
                        requested_version = int(popped_version[0])
                if filters:
                    query[param] = filters
            else:
                query[param] = param_value
        return query, requested_version

    def __init__(self, query_dict=None, data_dict=None):
        """
        Provide one of the 3 parameters depending on the format you have the query in.

        :param query_dict: a dict of query string parameters in the CKAN URL format - i.e. the
                           filters are split with colons and pipes etc
        :param data_dict: a dict of data dict parameters - i.e. the typical action data_dict format
        """
        if query_dict is not None:
            self.query, self.requested_version = self._parse_from_query_dict(query_dict)
        elif data_dict is not None:
            self.query, self.requested_version = self._parse_from_data_dict(data_dict)
        else:
            self.query = {}
            self.requested_version = None

        if self.requested_version is None:
            # default the requested time to now
            self.requested_version = int(time.time() * 1000)
        self.query_hash = self._generate_query_hash()

    def _generate_query_hash(self):
        """
        Create a unique hash for this query. To do this we have to ensure that the
        features like the order of filters is ignored to ensure that the meaning of the
        query is what we're capturing.

        :return: a unique hash of the query
        """
        query = {}
        for key, value in self.query.items():
            if key == 'filters':
                filters = {}
                for filter_field, filter_value in value.items():
                    # to ensure the order doesn't matter we have to convert everything to unicode
                    # and then sort it
                    filters[str(filter_field)] = sorted(map(str, filter_value))
                query['filters'] = filters
            else:
                query[str(key)] = str(value)

        # sort_keys=True is used otherwise the key ordering would change between python versions
        # and the hash wouldn't match even if the query was the same
        dumped_query = json.dumps(query, ensure_ascii=False, sort_keys=True).encode(
            'utf8'
        )
        return hashlib.sha1(dumped_query).hexdigest()

    def get_rounded_version(self, resource_id):
        """
        Round the requested version of this query down to the nearest actual version of
        the resource. See the versioned-search plugin for more details.

        :param resource_id: the id of the resource being searched
        :return: the rounded version or None if no versions are available for the given resource id
        """
        # first retrieve the rounded version to use
        data_dict = {'resource_id': resource_id, 'version': self.requested_version}
        return toolkit.get_action('datastore_get_rounded_version')({}, data_dict)

    def get_count(self, resource_id):
        """
        Retrieve the number of records matched by this query, resource id and version
        combination.

        :param resource_id: the resource id
        :return: an integer value
        """
        data_dict = copy.deepcopy(self.query)
        data_dict.update(
            {
                'resource_id': resource_id,
                # use the version parameter cause it's nicer than having to go in and modify the filters
                'version': self.get_rounded_version(resource_id),
                # we don't need the results, just the total
                'limit': 0,
            }
        )
        result = toolkit.get_action('datastore_search')({}, data_dict)
        return result['total']



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-query-dois
# Created by the Natural History Museum in London, UK
from typing import Optional

import json
from datetime import datetime

from ckan import model
from ckan.plugins import toolkit
from sqlalchemy import or_
from ckan.lib.helpers import link_to
from ckanext.query_dois.model import QueryDOI
from sqlalchemy.orm import Query


def render_filter_value(field, filter_value):
    """
    Renders the given filter value for the given field. This should be called for each
    filter value rather than by passing a list or filter values.

    :param field: the field name
    :param filter_value: the filter value for the field
    :return: the value to display
    """
    if field == '__geo__':
        return json.loads(filter_value)['type']
    else:
        return filter_value


def _make_all_resource_query(package_id: str) -> Optional[Query]:
    """
    Creates an SQL Alchemy query that can get all the QueryDOI entities associated with
    the resources of the given package.

    :param package_id:
    :return: None if the package doesn't exist or if the package has no resources,
             otherwise, returns an SQL Alchemy Query object
    """
    try:
        package = toolkit.get_action('package_show')({}, {'id': package_id})
    except toolkit.ObjectNotFound:
        return None
    ors = [QueryDOI.on_resource(resource['id']) for resource in package['resources']]
    if not ors:
        return None
    return model.Session.query(QueryDOI).filter(or_(*ors))


def get_most_recent_dois(package_id, number):
    """
    Retrieve the most recent DOIs that have been minted on queries against the resources
    in the given package.

    :param package_id: the package's ID
    :param number: the number of DOIs to return
    :return: a list of QueryDOI objects
    """
    query = _make_all_resource_query(package_id)
    if query is None:
        return []
    return list(query.order_by(QueryDOI.id.desc()).limit(number))


def get_doi_count(package_id: str) -> int:
    """
    Return the total count of DOIs created against the resources within the given
    package.

    :param package_id: the ID of the package
    :return: a number
    """
    query = _make_all_resource_query(package_id)
    return 0 if query is None else query.count()


# a tuple describing various ways of informing the user something happened a certain number of time
# units ago
time_resolutions = (
    (60, 'seceond', 1),
    (60 * 60, 'minute', 60),
    (60 * 60 * 24, 'hour', 60 * 60),
    (60 * 60 * 24 * 7, 'day', 60 * 60 * 24),
    (60 * 60 * 24 * 28, 'week', 60 * 60 * 24 * 7),
    (60 * 60 * 24 * 365, 'month', 60 * 60 * 24 * 28),
    (float('inf'), 'year', 60 * 60 * 24 * 365),
)


def get_time_ago_description(query_doi):
    """
    Given a QueryDOI object, return a short description of how long ago it was minted.
    The resolutions are described above in the time_resolutions tuple.

    :param query_doi: the QueryDOI object
    :return: a unicode string describing how long ago the DOI was minted
    """
    seconds = (datetime.now() - query_doi.timestamp).total_seconds()
    for limit, unit, divisor in time_resolutions:
        if seconds < limit:
            value = int(seconds / divisor)
            plural = 's' if value > 1 else ''
            return f'{value} {unit}{plural} ago'


def get_landing_page_url(query_doi):
    """
    Given a QueryDOI object, return the landing URL for it.

    :param query_doi: a QueryDOI object
    :return: the landing page URL
    """
    data_centre, identifier = query_doi.doi.split('/')
    return toolkit.url_for(
        'query_doi.landing_page', data_centre=data_centre, identifier=identifier
    )


def create_citation_text(
    query_doi,
    creation_timestamp,
    resource_name,
    package_title,
    package_doi=None,
    publisher=None,
    html=False,
):
    """
    Creates the citation text for the given query doi and the given additional related
    arguments.

    :param query_doi: the query's DOI, this should just be the prefix/suffix, e.g. 10.xxxx/xxxxxx,
                      not the full URL
    :param creation_timestamp: a datetime object representing the exact moment the DOI was created
    :param resource_name: the name of the resource the DOI references
    :param package_title: the title of the package the resource the DOI references is in
    :param package_doi: the DOI of the package, if there is one (defaults to None)
    :param publisher: the publisher to use in the citation (defaults to None in which case the
                      ckanext.query_dois.publisher config value is used
    :param html: whether to include a tags around URLs in the returned string. Defaults to False
                 which does not add a tags and therefore the returned string is just pure text
    :return: a citation string for the given query DOI and associated data
    """
    # default the publisher's value if needed
    if publisher is None:
        publisher = toolkit.config.get('ckanext.query_dois.publisher')

    # this is the citation's base form. This form is derived from the recommended RDA citation
    # format for evolving data when citing with a query. For more information see:
    # https://github.com/NaturalHistoryMuseum/ckanext-query-dois/issues/2
    citation_text = (
        '{publisher} ({year}). Data Portal Query on "{resource_name}" created at '
        '{creation_datetime} PID {query_doi}. Subset of "{dataset_name}" (dataset)'
    )

    # these are the parameters which will be used on the above string
    params = {
        'publisher': publisher,
        'year': creation_timestamp.year,
        'resource_name': resource_name,
        'creation_datetime': creation_timestamp,
        'query_doi': 'https://doi.org/{}'.format(query_doi),
        'dataset_name': package_title,
    }

    # if we have a DOI for the package, include it
    if package_doi is not None:
        citation_text += ' PID {dataset_doi}.'
        params['dataset_doi'] = f'https://doi.org/{package_doi}'

    if html:
        # there are currently two fields which should be converted to a tags
        for field in ('query_doi', 'dataset_doi'):
            doi_url = params.get(field, None)
            # the dataset_doi field is optional, hence the if here
            if doi_url is not None:
                params[field] = link_to(label=doi_url, url=doi_url, target='_blank')

    return citation_text.format(**params)


def create_multisearch_citation_text(query_doi, html=False):
    """
    Creates the citation text for a multisearch based DOI and returns it, either as a
    raw string or as a piece of html, if requested.

    :param query_doi: a query doi object
    :param html: whether to return a basic string or a piece of html
    :return: the citation text
    """
    publisher = toolkit.config.get('ckanext.query_dois.publisher')

    # this is the citation's base form. This form is derived from the recommended RDA citation
    # format for evolving data when citing with a query. For more information see:
    # https://github.com/NaturalHistoryMuseum/ckanext-query-dois/issues/2
    citation_text = (
        '{publisher} ({year}). Data Portal query on {resource_count} resources '
        'created at {creation_datetime} PID {query_doi}'
    )

    # these are the parameters which will be used on the above string
    params = {
        'publisher': publisher,
        'year': query_doi.timestamp.year,
        'resource_count': len(query_doi.get_resource_ids()),
        'creation_datetime': query_doi.timestamp,
    }

    doi_url = f'https://doi.org/{query_doi.doi}'
    if html:
        params['query_doi'] = link_to(label=doi_url, url=doi_url, target='_blank')
    else:
        params['query_doi'] = doi_url

    return citation_text.format(**params)


def pretty_print_query(query):
    """
    Does what you'd expect really.

    :param query: a query dict
    :return: a string of pretty json
    """
    return json.dumps(query, sort_keys=True, indent=2)
